/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <match_ip.h>

PLUGIN_DECLARE_STD_IMPORTS;

proc_matcher_add	add_matcher;

u_int16_t		fb_mask;

#define CHECK_IP(x) \
	(( NETWORK_LAYER(x).id == 0x0800 ) ? 1 : 0)

int fragbits_cmp(void *p1, void *p2)
{
	p_fragbits f1=(p_fragbits)p1;
	p_fragbits f2=(p_fragbits)p2;

	if ( (f1->mode==f2->mode) && (f1->fragbits==f2->fragbits) ) {
		return CMP_IDENTICAL;
	}

	return CMP_DIFFER;
}
	
int fragbits_validate(char *token, p_kv opts)
{
	p_fragbits priv;
	char *tmp;

	if ( !(priv=my_malloc(sizeof(fragbits_priv_t))) ) return NULL;
	memset(priv, 0, sizeof(priv));

	for (tmp=token; *tmp != '\0'; tmp++) {
		switch(*tmp) {
			case 'd':
			case 'D':
				priv->fragbits |= IP_DF;
				break;
				
			case 'm':
			case 'M':
				priv->fragbits |= IP_MF;
				break;
				
			case 'r':
			case 'R':
				priv->fragbits |= IP_RF;
				break;
			
			case '!':
				priv->mode = FBM_NOT;
				break;
				
			case '+':
				priv->mode = FBM_ANY;
				break;
				
			case '*':
				priv->mode = FBM_ALL;
				break;

			case ' ':
			case '\t':
			case ',':
				break;
			default:
				my_free(priv);
				return NULL;
		}
	}
	return priv;
}

int fragbits_match(void *priv, p_packet p, int negate)
{
	p_fragbits x=(p_fragbits)priv;
	u_int16_t fb=NETWORK_LAYER(p).h.ip->frag_off & fb_mask;

	if ( !CHECK_IP(p)) return 0;

	switch ( x->mode ) {
		case FBM_NORMAL:
			if ( fb == x->fragbits ) {
				MATCH_RET_POSITIVE;
			}else{
				break;
			}
		case FBM_NOT:
			if ( fb & x->fragbits == 0) {
				MATCH_RET_POSITIVE;
			}else{
				break;
			}
		case FBM_ALL:
			if ( fb & x->fragbits == x->fragbits ){
				MATCH_RET_POSITIVE;
			}else{
				break;
			}
		case FBM_ANY:
			if ( fb & x->fragbits != 0){
				MATCH_RET_POSITIVE;
			}else{
				break;
			}

		default:
			return 0;
			
	}
	
	MATCH_RET_NEGATIVE;
}
	
int tos_match(void *priv, p_packet p, int negate)
{
	int tos;
	p_range_priv x=(p_range_priv)priv;
	
	/* Make sure it's IP */
	if (!CHECK_IP(p)) return 0;

	/* Clamp values to maximum */
	if ( x->min > 65535 ) x->min=65535;
	if ( x->max > 65535 ) x->max=65535;
	
	tos = ntohs(NETWORK_LAYER(p).h.ip->tos);
	
	if ( (tos >= x->min) && (tos <= x->max) ) {
		MATCH_RET_POSITIVE;
	}else{
		MATCH_RET_NEGATIVE;
	}
}

int ipid_match(void *priv, p_packet p, int negate)
{
	int ipid;
	p_range_priv x=(p_range_priv)priv;
	
	/* Make sure it's IP */
	if (!CHECK_IP(p)) return 0;

	/* Clamp values to maximum */
	if ( x->min > 65535 ) x->min=65535;
	if ( x->max > 65535 ) x->max=65535;
	
	ipid = ntohs(NETWORK_LAYER(p).h.ip->id);
	
	if ( (ipid >= x->min) && (ipid <= x->max) ) {
		MATCH_RET_POSITIVE;
	}else{
		MATCH_RET_NEGATIVE;
	}
}


int ttl_match(void *priv, p_packet p, int negate)
{
	int ttl;
	p_range_priv x=(p_range_priv)priv;
	
	/* Make sure it's IP */
	if (!CHECK_IP(p)) return 0;

	ttl=NETWORK_LAYER(p).h.ip->ttl;

	/* Clamp values to maximum */
	if ( x->min > 255 ) x->min=255;
	if ( x->max > 255 ) x->max=255;
	
	if ( (ttl >= x->min) && (ttl <= x->max) ) {
		MATCH_RET_POSITIVE;
	}else{
		MATCH_RET_NEGATIVE;
	}
}
uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	PLUGIN_CHECK_PARAMS(in);
	PLUGIN_IMPORT_STD;

	add_matcher=in->add_matcher;

	PLUGIN_NAME("match/ip");
	PLUGIN_DESC("Some generic IP based matchers");
	PLUGIN_VERS(1,0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	fb_mask=htons(0xE000);
	
	add_matcher("TTL",
		in->templates.t_validate_range,
		in->templates.t_cleanup_ptr,
		ttl_match,
		in->templates.t_cmp_range);
	
	add_matcher("TOS",
		in->templates.t_validate_range,
		in->templates.t_cleanup_ptr,
		tos_match,
		in->templates.t_cmp_range);
	
	add_matcher("ID",
		in->templates.t_validate_range,
		in->templates.t_cleanup_ptr,
		ipid_match,
		in->templates.t_cmp_range);
	
	add_matcher("FRAGBITS",
		fragbits_validate,
		in->templates.t_cleanup_ptr,
		fragbits_match,
		fragbits_cmp);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	return PLUGIN_ERR_OK;
}
