/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <match_str.h>
#include <s_util.h>

PLUGIN_DECLARE_STD_IMPORTS;

proc_matcher_add	add_matcher;
proc_list_append	list_append;
p_str_seg		segments=NULL;

int string_cmp(void *p1, void *p2)
{
	p_str_priv s1=(p_str_priv)p1;
	p_str_priv s2=(p_str_priv)p2;

	if ( 
		(s1->len==s2->len)  &&
		(s1->offset==s2->offset) &&
		(s1->depth==s2->depth) &&
		(s1->nocase==s2->nocase) &&
		memcmp(s1->str, s2->str, s1->len)==0
	) {
		return CMP_IDENTICAL;
	}

	return CMP_DIFFER;
}

/* Stitch all the segments together */
void string_construct(p_str_priv priv)
{
	p_str_seg tmp;
	void *cur;
	
	if ( !priv ) return;
	
	/* Count the length of the string */
	for ( tmp=segments, priv->len=0; tmp; tmp=(p_str_seg)tmp->l.next) {
		priv->len+=tmp->len;
	}

	/* Allocate the space */
	if ( !(priv->str=my_malloc(priv->len+1)) ) return;
	memset(priv->str, 0, priv->len+1);

	/* Copy it all in */
	cur=priv->str;
	for ( tmp=segments; tmp; tmp=(p_str_seg)tmp->l.next) {
		memcpy(cur, tmp->seg, tmp->len);
		cur+=tmp->len;
	}
}

void string_kill_segs(void)
{
	p_str_seg foo, bar;

	for(foo=segments; foo; ){
		bar=foo;
		foo=(p_c_capdev)bar->l.next;
		if ( bar->seg ) my_free(bar->seg);
		my_free(bar);
	}
	segments=NULL;
}

int string_add_seg(int bin, char *str)
{
	p_str_seg	tmp;
	char		*b;
	int		bi;
	char		byte[3];
	uint		digit;
	uint		dep=0;

	/* print_out("FRAG: %s : %s" CRLF, (bin)?"bin":"str", str); */
	
	/* Allocate the new segment */
	if ( !(tmp=my_malloc(sizeof(str_seg_t))) ){
		print_out(PRINT_ERROR "Couldn't allocate string segment" CRLF);
		return 0;
	}

	if ( bin ) {
		/* Handle binary segments */
		
		tmp->len=0;
		for(b=str; *b != '\0'; b++) {
			if ( isxdigit(*b) ) tmp->len++;
		}

		if ( tmp->len % 2 != 0 ) {
			print_out(PRINT_ERROR "Binary data in string "
				"adds up to %i _and_a_half_ bytes!" CRLF,
			       	tmp->len/2);
			my_free(tmp);
			return 0;
		}
		
		tmp->len/=2;

		if ( !(tmp->seg=my_malloc(tmp->len)) ) {
			my_free(tmp);
			print_raw(PRINT_ERROR "Couldn't allocate string" CRLF);
			return 0;
		}

		byte[3]='\0';
		bi=0;
		for(b=str; *b != '\0'; b++) {
			 if ( isxdigit(*b) ) {
				byte[bi]=*b;
				bi++;
				
				if ( bi == 2 ) {
					bi=0;
					digit=strtoul((char *)&byte, NULL, 16);

					*((char *)(tmp->seg+dep))=(char)digit;
					dep++;
				}
			 }
		}
	}else{
		/* Handle plain old bits of text */
		tmp->len=strlen(str);
		tmp->seg=my_strdup(str);
		if ( !tmp->seg ) {
			print_raw(PRINT_ERROR "Couldn't copy string segment");
			my_free(tmp);
			return 0;
		}
	}
	
	segments=list_append(tmp, segments);
	return 1;
}

int string_get_segs(char *str)
{
	char *tmp;
	char *startseg;
	int inbin=0;
	
	if ( !str ) return 0;

	startseg=str;
	for(tmp=str; *tmp; tmp++) {
		switch (*tmp) {
			case '|':
				*tmp='\0';
				if ( !string_add_seg(inbin, startseg) ) return 0;
				inbin=(inbin) ? 0 : 1 ;
				startseg=tmp+1;
				break;
				
			default:
				break;

		}
	}
	return string_add_seg(inbin, startseg);
}

void *string_validate(char *token, p_kv opts)
{
	p_str_priv priv;
	p_kv k;
#ifdef _STR_DEBUG
	char *tmp;
#endif
	
	if ( !token ) return NULL;
	if ( !(priv=my_malloc(sizeof(str_priv_t))) ) return NULL;
	
	priv->depth=0;
	priv->offset=0;
	
	for ( k=opts; k; k=(p_kv)k->l.next) {
		if ( !strcasecmp(k->key, "DEPTH") ) {
			priv->depth = atoi(k->val);
			if ( !priv->depth ) {
				print_out(PRINT_ERROR "Depth is 0??" CRLF);
				my_free(priv);
				return NULL;
			}
			continue;
		}
		
		if ( !strcasecmp(k->key, "OFFSET") ) {
			priv->offset = atoi(k->val);
			continue;
		}
		
		if ( !strcasecmp(k->key, "NOCASE") ) {
			priv->nocase = 1;
			continue;
		}
	}

#ifdef _STR_DEBUG
	print_out("str('%s'): ", token);
#endif
	
	if ( !string_get_segs(token) ) {
		string_kill_segs();
		my_free(priv);
		return NULL;
	}
	
	string_construct(priv);

#ifdef _STR_DEBUG
	for(tmp=priv->str; *tmp; tmp++) {
		print_out("%.2X ", *tmp);
	}print_raw(CRLF);
#endif

	string_kill_segs();
	
	return priv;
}

void string_cleanup(void *p)
{
	p_str_priv priv=p;
	
	if ( !priv ) return;

	if ( priv->str ) my_free(priv->str);
	
	my_free(priv);
}

int string_match(void *priv, p_packet p, int negate)
{
	p_str_priv sp=priv;
	char *str, *end;
	int hlen=0, nlen=0, slen=0;
	int i;
	
	if ( !(str=APPLICATION_LAYER(p).h.raw) ) return 0;

	hlen = ( p->caplen - ((unsigned long)str - (unsigned long)p->data) );
	nlen = strlen(sp->str);
	
	
	/* Depth ovverides slen */
	if ( sp->depth ) {
		slen=sp->depth;
	}else{
		slen=(hlen-nlen);
		str+=sp->offset;
		slen-=sp->offset;
	}
	
	if ( slen <= 0 ) return 0;
	
	end = p->data + p->caplen - nlen;
	
	for (i=0; i<slen; i++, str++) {
		if ( str > end ) break;
		if ( memcmp(sp->str, str, nlen) == 0 ) {
			MATCH_RET_POSITIVE;
		}
	}

	MATCH_RET_NEGATIVE;
}

uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	PLUGIN_CHECK_PARAMS(in);
	PLUGIN_IMPORT_STD;

	add_matcher=in->add_matcher;
	list_append=in->list_append;
	
	PLUGIN_NAME("match/string");
	PLUGIN_DESC("Matches strings in packets");
	PLUGIN_VERS(1,0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	add_matcher("CONTENT",
		string_validate,
		string_cleanup,
		string_match,
		string_cmp);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	return PLUGIN_ERR_OK;
}
